@@ -21,6 +21,7 @@ from bluetooth_sensor_state_data import BluetoothData
 from Cryptodome.Cipher import AES
 from home_assistant_bluetooth import BluetoothServiceInfo
 from sensor_state_data import (
+    BaseDeviceClass,
     BinarySensorDeviceClass,
     SensorLibrary,
     SensorUpdate,
@@ -34,6 +35,7 @@ from .const import (
     TIMEOUT_1DAY,
 )
 from .devices import DEVICE_TYPES
+from .events import EventDeviceKeys
 
 _LOGGER = logging.getLogger(__name__)
 
@@ -50,6 +52,25 @@ class EncryptionScheme(Enum):
     MIBEACON_4_5 = "mibeacon_4_5"
 
 
+class ExtendedBinarySensorDeviceClass(BaseDeviceClass):
+    """Device class for additional binary sensors (compared to sensor-state-data)."""
+
+    # On means door left open, Off means door closed
+    DEVICE_FORCIBLY_REMOVED = "device_forcibly_removed"
+
+    # On means door left open, Off means door closed
+    DOOR_LEFT_OPEN = "door_left_open"
+
+    # On means door stuck, Off means clear
+    DOOR_STUCK = "door_stuck"
+
+    # On means door someone knocking on the door, Off means no knocking
+    KNOCK_ON_THE_DOOR = "knock_on_the_door"
+
+    # On means door pried, Off means door not pried
+    PRY_THE_DOOR = "pry_the_door"
+
+
 def to_mac(addr: bytes) -> str:
     """Return formatted MAC address"""
     return ":".join(f"{i:02X}" for i in addr)
@@ -137,7 +158,14 @@ def obj0003(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
 ) -> dict[str, Any]:
     """Motion"""
-    return {"motion": xobj[0], "motion timer": xobj[0]}
+    # 0x0003 is only used by MUE4094RT, which does not send motion clear.
+    # This object is therefore added as event (motion detected).
+    device.fire_event(
+        key=EventDeviceKeys.MOTION,
+        event_type="motion_detected",
+        event_properties=None,
+    )
+    return {}
 
 
 def obj0006(
@@ -187,40 +215,86 @@ def obj0007(
     """Door"""
     door_byte = xobj[0]
     if door_byte == 0x00:
-        action = "open the door"
-        door = 1
+        # open the door
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.DOOR, True)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DOOR_STUCK,
+            native_value=False,  # reset door stuck
+            device_class=ExtendedBinarySensorDeviceClass.DOOR_STUCK,
+            name="Door stuck",
+        )
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.KNOCK_ON_THE_DOOR,
+            native_value=False,  # reset knock on the door
+            device_class=ExtendedBinarySensorDeviceClass.KNOCK_ON_THE_DOOR,
+            name="Knock on the door",
+        )
     elif door_byte == 0x01:
-        action = "close the door"
-        door = 0
+        # close the door
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.DOOR, False)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            native_value=False,  # reset door left open
+            device_class=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            name="Door left open",
+        )
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.PRY_THE_DOOR,
+            native_value=False,  # reset pry the door
+            device_class=ExtendedBinarySensorDeviceClass.PRY_THE_DOOR,
+            name="Pry the door",
+        )
     elif door_byte == 0x02:
-        action = "timeout, not closed"
-        door = 1
+        # timeout, not closed
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.DOOR, True)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            native_value=True,
+            device_class=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            name="Door left open",
+        )
     elif door_byte == 0x03:
-        action = "knock on the door"
-        door = 0
+        # knock on the door
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.DOOR, False)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.KNOCK_ON_THE_DOOR,
+            native_value=True,
+            device_class=ExtendedBinarySensorDeviceClass.KNOCK_ON_THE_DOOR,
+            name="Knock on the door",
+        )
     elif door_byte == 0x04:
-        action = "pry the door"
-        door = 1
+        # pry the door
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.DOOR, True)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.PRY_THE_DOOR,
+            native_value=True,
+            device_class=ExtendedBinarySensorDeviceClass.PRY_THE_DOOR,
+            name="Pry the door",
+        )
     elif door_byte == 0x05:
-        action = "door stuck"
-        door = 0
-    else:
-        return {}
-    return {"door": door, "door action": action}
+        # door stuck
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.DOOR, False)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DOOR_STUCK,
+            native_value=True,
+            device_class=ExtendedBinarySensorDeviceClass.DOOR_STUCK,
+            name="Door stuck",
+        )
+    return {}
 
 
 def obj0008(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
 ) -> dict[str, Any]:
     """armed away"""
-    returnData: dict[str, Any] = {}
+    return_data: dict[str, Any] = {}
     value = xobj[0] ^ 1
-    returnData.update({"armed away": value})
+    return_data.update({"armed away": value})
     if len(xobj) == 5:
         timestamp = datetime.datetime.utcfromtimestamp(
             int.from_bytes(xobj[1:], "little")
         ).isoformat()
-        returnData.update({"timestamp": timestamp})
+        return_data.update({"timestamp": timestamp})
     # Lift up door handle outside the door sends this event from DSL-C08.
     if device_type == "DSL-C08":
         return {
@@ -232,7 +306,7 @@ def obj0008(
             "key id": None,
             "timestamp": None,
         }
-    return returnData
+    return return_data
 
 
 def obj0010(
@@ -306,22 +380,17 @@ def obj000f(
     """Moving with light"""
     if len(xobj) == 3:
         (illum,) = LIGHT_STRUCT.unpack(xobj + b"\x00")
-
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.MOTION, True)
         if device_type in ["MJYD02YL", "RTCGQ02LM"]:
             # MJYD02YL:  1 - moving no light, 100 - moving with light
             # RTCGQ02LM: 0 - moving no light, 256 - moving with light
             device.update_predefined_binary_sensor(
                 BinarySensorDeviceClass.LIGHT, bool(illum >= 100)
             )
-            return {"motion": 1, "motion timer": 1}
         elif device_type == "CGPR1":
             # CGPR1:     moving, value is illumination in lux
             device.update_predefined_sensor(SensorLibrary.LIGHT__LIGHT_LUX, illum)
-            return {"motion": 1, "motion timer": 1}
-        else:
-            return {}
-    else:
-        return {}
+    return {}
 
 
 def obj1001(
@@ -540,8 +609,6 @@ def obj1007(
         elif device_type in ["HHCCJCY01", "GCLS002"]:
             # illumination in lux
             device.update_predefined_sensor(SensorLibrary.LIGHT__LIGHT_LUX, illum)
-        else:
-            return {}
     return {}
 
 
@@ -573,9 +640,7 @@ def obj1010(
             SensorLibrary.FORMALDEHYDE__CONCENTRATION_MILLIGRAMS_PER_CUBIC_METER,
             fmdh / 100,
         )
-        return {}
-    else:
-        return {}
+    return {}
 
 
 def obj1012(
@@ -623,20 +688,24 @@ def obj1017(
     if len(xobj) == 4:
         (no_motion_time,) = M_STRUCT.unpack(xobj)
         # seconds since last motion detected message
-        # 0 = motion detected
-        return {
-            "motion": 1 if no_motion_time == 0 else 0,
-            "no motion time": no_motion_time,
-        }
-    else:
-        return {}
+        # 0x1017 is send 3 seconds after 0x000f, 5 seconds arter 0x1007
+        # and at 60, 120, 300, 600, 1200 and 1800 seconds after last motion.
+        # Anything <= 30 seconds is regarded motion detected in the MiHome app.
+        if no_motion_time <= 30:
+            device.update_predefined_binary_sensor(BinarySensorDeviceClass.MOTION, True)
+        else:
+            device.update_predefined_binary_sensor(
+                BinarySensorDeviceClass.MOTION, False
+            )
+    return {}
 
 
 def obj1018(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
 ) -> dict[str, Any]:
     """Light intensity"""
-    return {"light": xobj[0]}
+    device.update_predefined_binary_sensor(BinarySensorDeviceClass.LIGHT, bool(xobj[0]))
+    return {}
 
 
 def obj1019(
@@ -645,21 +714,32 @@ def obj1019(
     """Door/Window sensor"""
     open_obj = xobj[0]
     if open_obj == 0:
-        opening = 1
-        status = "opened"
+        # opened
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.OPENING, True)
     elif open_obj == 1:
-        opening = 0
-        status = "closed"
+        # closed
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.OPENING, False)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            native_value=False,  # reset door left open
+            device_class=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            name="Door left open",
+        )
     elif open_obj == 2:
-        opening = 1
-        status = "closing timeout"
+        # closing timeout
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.OPENING, True)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            native_value=True,
+            device_class=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            name="Door left open",
+        )
     elif open_obj == 3:
-        opening = 1
-        status = "device reset"
+        # device reset (not implemented)
+        return {}
     else:
-        opening = 0
-        status = None
-    return {"opening": opening, "status": status}
+        return {}
+    return {}
 
 
 def obj100a(
@@ -701,6 +781,15 @@ def obj100e(
     return {}
 
 
+def obj101b(
+    xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
+) -> dict[str, Any]:
+    """Timeout no movement"""
+    # https://iot.mi.com/new/doc/accesses/direct-access/embedded-development/ble/object-definition#%E9%80%9A%E7%94%A8%E5%B1%9E%E6%80%A7
+    device.update_predefined_binary_sensor(BinarySensorDeviceClass.MOTION, False)
+    return {}
+
+
 def obj2000(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
 ) -> dict[str, Any]:
@@ -721,8 +810,8 @@ def obj2000(
     return {}
 
 
-# The following data objects are device specific. For now only
-#  added for LYWSD02MMC, XMWSDJ04MMC, XMWXKG01YL, LINPTECH MS1BB(MI), HS1BB(MI)
+# The following data objects are device specific. For now only added for
+# LYWSD02MMC, MJWSD05MMC, XMWSDJ04MMC, XMWXKG01YL, LINPTECH MS1BB(MI), HS1BB(MI), K9BB
 # https://miot-spec.org/miot-spec-v2/instances?status=all
 def obj4803(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
@@ -739,20 +828,31 @@ def obj4804(
     opening_state = xobj[0]
     # State of the door/window, used in combination with obj4a12
     if opening_state == 1:
-        opening = 1
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.OPENING, True)
     elif opening_state == 2:
-        opening = 0
-    else:
-        return {}
-    return {"opening": opening}
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.OPENING, False)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            native_value=False,  # reset door left open
+            device_class=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            name="Door left open",
+        )
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DEVICE_FORCIBLY_REMOVED,
+            native_value=False,  # reset device forcibly removed
+            device_class=ExtendedBinarySensorDeviceClass.DEVICE_FORCIBLY_REMOVED,
+            name="Device forcibly removed",
+        )
+    return {}
 
 
 def obj4805(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
 ) -> dict[str, Any]:
     """Illuminance in lux"""
-    (illu,) = struct.unpack("f", xobj)
-    return {"illuminance": illu}
+    (illum,) = struct.unpack("f", xobj)
+    device.update_predefined_sensor(SensorLibrary.LIGHT__LIGHT_LUX, illum)
+    return {}
 
 
 def obj4818(
@@ -763,12 +863,15 @@ def obj4818(
         (no_motion_time,) = struct.unpack("<H", xobj)
         # seconds since last motion detected message
         # 0 = motion detected
-        return {
-            "motion": 1 if no_motion_time == 0 else 0,
-            "no motion time": no_motion_time,
-        }
-    else:
-        return {}
+        # also send at 60, 120, 300, 600, 1200 and 1800 seconds after last motion.
+        # Anything <= 30 seconds is regarded motion detected in the MiHome app.
+        if no_motion_time <= 30:
+            device.update_predefined_binary_sensor(BinarySensorDeviceClass.MOTION, True)
+        else:
+            device.update_predefined_binary_sensor(
+                BinarySensorDeviceClass.MOTION, False
+            )
+    return {}
 
 
 def obj4a01(
@@ -783,19 +886,26 @@ def obj4a08(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
 ) -> dict[str, Any]:
     """Motion detected with Illuminance in lux"""
-    (illu,) = struct.unpack("f", xobj)
-    return {"motion": 1, "motion timer": 1, "illuminance": illu}
+    (illum,) = struct.unpack("f", xobj)
+    device.update_predefined_binary_sensor(BinarySensorDeviceClass.MOTION, True)
+    device.update_predefined_sensor(SensorLibrary.LIGHT__LIGHT_LUX, illum)
+    return {}
 
 
 def obj4a0f(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
 ) -> dict[str, Any]:
-    """Door/window broken open"""
+    """Device forcibly removed"""
     dev_forced = xobj[0]
     if dev_forced == 1:
-        return {"opening": 1, "status": "door/window broken open"}
-    else:
-        return {}
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.OPENING, True)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DEVICE_FORCIBLY_REMOVED,
+            native_value=True,
+            device_class=ExtendedBinarySensorDeviceClass.DEVICE_FORCIBLY_REMOVED,
+            name="Device forcibly removed",
+        )
+    return {}
 
 
 def obj4a12(
@@ -805,12 +915,22 @@ def obj4a12(
     opening_state = xobj[0]
     # Opening event, used in combination with obj4804
     if opening_state == 1:
-        opening = 1
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.OPENING, True)
     elif opening_state == 2:
-        opening = 0
-    else:
-        return {}
-    return {"opening": opening}
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.OPENING, False)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            native_value=False,
+            device_class=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            name="Door left open",
+        )
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DEVICE_FORCIBLY_REMOVED,
+            native_value=False,  # reset device forcibly removed
+            device_class=ExtendedBinarySensorDeviceClass.DEVICE_FORCIBLY_REMOVED,
+            name="Device forcibly removed",
+        )
+    return {}
 
 
 def obj4a13(
@@ -820,18 +940,22 @@ def obj4a13(
     click = xobj[0]
     if click == 1:
         return {"button": "toggle"}
-    else:
-        return {}
+    return {}
 
 
 def obj4a1a(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
 ) -> dict[str, Any]:
-    """Door Not Closed"""
+    """Door left open"""
     if xobj[0] == 1:
-        return {"opening": 1, "status": "door not closed"}
-    else:
-        return {}
+        device.update_predefined_binary_sensor(BinarySensorDeviceClass.OPENING, True)
+        device.update_binary_sensor(
+            key=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            native_value=False,
+            device_class=ExtendedBinarySensorDeviceClass.DOOR_LEFT_OPEN,
+            name="Door left open",
+        )
+    return {}
 
 
 def obj4c01(
@@ -887,15 +1011,44 @@ def obj4e0c(
     result: dict[str, Any] = {}
 
     click = xobj[0]
-
-    if click == 1:
-        result["two_btn_switch_left"] = "toggle"
-    elif click == 2:
-        result["two_btn_switch_right"] = "toggle"
-    elif click == 3:
-        result["two_btn_switch_left"] = "toggle"
-        result["two_btn_switch_right"] = "toggle"
-
+    if device_type == "XMWXKG01YL":
+        if click == 1:
+            result = {
+                "two btn switch left": "toggle",
+                "button switch": "single press",
+            }
+        elif click == 2:
+            result = {
+                "two btn switch right": "toggle",
+                "button switch": "single press",
+            }
+        elif click == 3:
+            result = {
+                "two btn switch left": "toggle",
+                "two btn switch right": "toggle",
+                "button switch": "single press",
+            }
+    elif device_type == "K9BB-1BTN":
+        if click == 1:
+            result = {
+                "one btn switch": "toggle",
+                "button switch": "single press",
+            }
+        elif click == 8:
+            result = {
+                "one btn switch": "toggle",
+                "button switch": "long press",
+            }
+        elif click == 15:
+            result = {
+                "one btn switch": "toggle",
+                "button switch": "double press",
+            }
+    elif device_type == "XMWXKG01LM":
+        result = {
+            "one btn switch": "toggle",
+            "button switch": "single press",
+        }
     return result
 
 
@@ -903,48 +1056,71 @@ def obj4e0d(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
 ) -> dict[str, Any]:
     """Double Click"""
+    result: dict[str, Any] = {}
+
     click = xobj[0]
-    btn_switch_press_type: str | None = "double press"
-    two_btn_switch_left = None
-    two_btn_switch_right = None
-    if click == 1:
-        two_btn_switch_left = "toggle"
-    elif click == 2:
-        two_btn_switch_right = "toggle"
-    elif click == 3:
-        two_btn_switch_left = "toggle"
-        two_btn_switch_right = "toggle"
-    else:
-        btn_switch_press_type = None
-    return {
-        "two btn switch left": two_btn_switch_left,
-        "two btn switch right": two_btn_switch_right,
-        "button switch": btn_switch_press_type,
-    }
+    if device_type == "XMWXKG01YL":
+        if click == 1:
+            result = {
+                "two btn switch left": "toggle",
+                "button switch": "double press",
+            }
+        elif click == 2:
+            result = {
+                "two btn switch right": "toggle",
+                "button switch": "double press",
+            }
+        elif click == 3:
+            result = {
+                "two btn switch left": "toggle",
+                "two btn switch right": "toggle",
+                "button switch": "double press",
+            }
+    elif device_type == "XMWXKG01LM":
+        result = {
+            "one btn switch": "toggle",
+            "button switch": "double press",
+        }
+    return result
 
 
 def obj4e0e(
     xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
 ) -> dict[str, Any]:
     """Long Press"""
+    result: dict[str, Any] = {}
+
     click = xobj[0]
-    btn_switch_press_type: str | None = "long press"
-    two_btn_switch_left = None
-    two_btn_switch_right = None
-    if click == 1:
-        two_btn_switch_left = "toggle"
-    elif click == 2:
-        two_btn_switch_right = "toggle"
-    elif click == 3:
-        two_btn_switch_left = "toggle"
-        two_btn_switch_right = "toggle"
-    else:
-        btn_switch_press_type = None
-    return {
-        "two btn switch left": two_btn_switch_left,
-        "two btn switch right": two_btn_switch_right,
-        "button switch": btn_switch_press_type,
-    }
+    if device_type == "XMWXKG01YL":
+        if click == 1:
+            result = {
+                "two btn switch left": "toggle",
+                "button switch": "long press",
+            }
+        elif click == 2:
+            result = {
+                "two btn switch right": "toggle",
+                "button switch": "long press",
+            }
+        elif click == 3:
+            result = {
+                "two btn switch left": "toggle",
+                "two btn switch right": "toggle",
+                "button switch": "long press",
+            }
+    elif device_type == "XMWXKG01LM":
+        result = {
+            "one btn switch": "toggle",
+            "button switch": "long press",
+        }
+    return result
+
+
+def obj4e1c(
+    xobj: bytes, device: XiaomiBluetoothDeviceData, device_type: str
+) -> dict[str, Any]:
+    """Device reset"""
+    return {"device reset": True}
 
 
 # Dataobject dictionary
@@ -975,6 +1151,7 @@ xiaomi_dataobject_dict = {
     0x100A: obj100a,
     0x100D: obj100d,
     0x100E: obj100e,
+    0x101B: obj101b,
     0x2000: obj2000,
     0x4803: obj4803,
     0x4804: obj4804,
@@ -991,6 +1168,7 @@ xiaomi_dataobject_dict = {
     0x4C03: obj4c03,
     0x4C08: obj4c08,
     0x4C14: obj4c14,
+    0x4E1C: obj4e1c,
     0x4E0C: obj4e0c,
     0x4E0D: obj4e0d,
     0x4E0E: obj4e0e,
